##
## clirc -- An IRC client in Python.
## main.py - The main program, separated to a module to enable reloading.
##
## Copyleft 1997-1999 Teemu Kalvas <chery@s2.org>
##

import os
import select
import signal
import string
import sys
import time
import traceback

import reloadable

import b64
import config
import crc32
import crypt
import editfield
import idea
import rnd
import screen
import server
import statline
import user
import window

def child_killed(a, b):
    p.child_killed = 1

signal.signal(signal.SIGCHLD, child_killed)

class P(reloadable.Reloadable):
    def __init__(self):
	global p
	p = self
	self.post_reload()

    def post_reload(self):
	self.ensure_defaults([('quit', 0),
			      ('wait', []),
			      ('screen', lambda: screen.Screen()),
			      ('rand', int(time.time())),
			      ('in_exception', 0),
			      ('child_killed', 0),
			      ('dcc_count', 0),
			      ('dcc_hash', {}),
			      ('dcc_pid', {}),
			      ('dcc_receiving', {}),
			      ('dcc_sending', {})])

    def wait_add(self, time, func, context):
	i = 0
	for entry in self.wait:
	    if entry[0] > time:
		self.wait.insert(i, (time, func, context))
		return
	    i = i + 1
	self.wait.append((time, func, context))

    def initialize(p):
	p.servers = [user.User()]
	for serv in config.want_servers:
	    srv = server.Server(p, serv)
	    if srv.connected:
		p.servers.append(srv)

    def inner(p):
	try:
	    if p.child_killed:
		p.child_killed = 0
		pid, status = os.wait()
		if p.dcc_pid.has_key(pid):
		    number, direction = p.dcc_pid[pid]
		    del p.dcc_pid[pid]
		    if direction:
			del p.dcc_sending[number]
			p.screen.current.log('*** DCC send #' + repr(number)
					     + ' finished')
		    else:
			del p.dcc_receiving[number]
			p.screen.current.log('*** DCC receive #' + repr(number)
					     + ' finished')
		## It is unclear whether we should do this so we make sure.
		signal.signal(signal.SIGCHLD, child_killed)
	    if p.screen.touched:
		p.servers[0].out = [1]
	    rl = map(lambda x: x.sock,
		     filter(lambda x: x.connected, p.servers))
	    sl = filter(lambda x: len(x.out) > 0 and x.connected, p.servers)
	    wl = map(lambda x: x.sock, sl)
	    if len(p.wait) > 0:
		timeout = max(p.wait[0][0] - time.time(), 0.0)
	    else:
		timeout = 3600.0
	    rr, wr = select.select(rl, wl, [], timeout)[:2]
	    if len(rr) > 0:
		for serv in p.servers[:]:
		    if serv.sock == rr[0]:
			serv.read(p)
			if not serv.connected:
			    if serv.quitting:
				serv.clean(p)
			    else:
				p.screen.touched = p.screen.touched | 16
			del rr[0]
			if len(rr) == 0:
			    break
	    elif len(wr) > 0:
		for serv in p.servers:
		    if serv.sock == wr[0]:
			if serv.connected:
			    serv.write(p)
			else:
			    if serv.quitting:
				serv.clean(p)
			    else:
				p.screen.touched = p.screen.touched | 16
			del wr[0]
			if len(wr) == 0:
			    break
	    while len(p.wait) > 0 and p.wait[0][0] <= time.time():
		p.wait[0][1](p, p.wait[0][2])
		del p.wait[0]
		p.in_exception = 0
	except select.error:
	    pass
	except:
	    if p.in_exception:
		dump = sys.exc_type, sys.exc_value, sys.exc_traceback
		screen.clean()
		traceback.print_exception(dump[0], dump[1], dump[2])
		sys.exit(1)
	    p.in_exception = 1
	    slist = traceback.format_exception(
		sys.exc_type, sys.exc_value, sys.exc_traceback)
	    for s in slist:
		for l in string.split(s[:-1], '\n'):
		    p.screen.current.log('+++ ' + l)

    def do_reload(p):
	p.post_reload()
	try:
	    reload(b64)
	    reload(crc32)
	    reload(crypt)
	    reload(editfield)
	    reload(rnd)
	    reload(screen)
	    reload(server)
	    reload(statline)
	    reload(user)
	    reload(window)
	    reload(config)
	    for old, new in [(p.screen.__class__, screen.Screen),
			     (p.screen.edit_field.__class__,
			      editfield.EditField),
			     (p.screen.stat_line.__class__,
			      statline.StatLine),
			     (p.screen.current.__class__, window.Window)]:
		for attr in new.__dict__.keys():
		    old.__dict__[attr] = new.__dict__[attr]
	    p.servers[0].__class__ = user.User
	    for s in p.servers[1:]:
		s.__class__ = server.Server
	    for s in p.servers:
		s.post_reload()
	except:
	    slist = traceback.format_exception(
		sys.exc_type, sys.exc_value, sys.exc_traceback)
	    for s in slist:
		for l in string.split(s[:-1], '\n'):
		    p.screen.current.log('+++ ' + l)

## End. ##
